from __future__ import annotations

import datetime
import json
import os
from contextlib import suppress
from typing import TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from typing import Any

    from adaptive_scheduler.scheduler import BaseScheduler

    from .database_manager import DatabaseManager


def _get_infos(fname: str, *, only_last: bool = True) -> list[dict[str, Any]]:
    status_lines: list[dict[str, Any]] = []
    with open(fname, encoding="utf-8") as f:  # noqa: PTH123
        lines = f.readlines()
        for line in reversed(lines):
            with suppress(Exception):
                info = json.loads(line)
                if info["event"] == "current status":
                    status_lines.append(info)
                    if only_last:
                        return status_lines
        return status_lines


def parse_log_files(
    database_manager: DatabaseManager,
    scheduler: BaseScheduler,
    *,
    only_last: bool = True,
) -> pd.DataFrame:
    """Parse the log-files and convert it to a `~pandas.core.frame.DataFrame`.

    Parameters
    ----------
    job_names
        List of job names.
    database_manager
        A `DatabaseManager` instance.
    scheduler
        A scheduler instance from `adaptive_scheduler.scheduler`.
    only_last
        Only look use the last printed status message.

    Returns
    -------
    `~pandas.core.frame.DataFrame`

    """
    _queue = scheduler.queue()
    database_manager.update(_queue)

    infos = []
    for entry in database_manager.as_dicts():
        log_fname = entry["log_fname"]
        if log_fname is None or not os.path.exists(log_fname):  # noqa: PTH110
            continue
        for info_dict in _get_infos(log_fname, only_last=only_last):
            info_dict.pop("event")  # this is always "current status"
            info_dict["timestamp"] = datetime.datetime.strptime(  # noqa: DTZ007
                info_dict["timestamp"],
                "%Y-%m-%d %H:%M.%S",
            )
            info_dict["elapsed_time"] = pd.to_timedelta(info_dict["elapsed_time"])
            info_dict.update(entry)
            if "npoints" in info_dict:  # ensure it is an int, xref: #223
                info_dict["npoints"] = int(info_dict["npoints"])
            infos.append(info_dict)

    for info_dict in infos:
        info_from_queue = _queue.get(info_dict["job_id"])
        if info_from_queue is None:
            continue
        info_dict["state"] = info_from_queue["state"]
        info_dict["job_name"] = info_from_queue["job_name"]

    return pd.DataFrame(infos)



"""Tests for `adaptive_scheduler.utils`."""

from __future__ import annotations

import platform
import sys
import time
import typing
from concurrent.futures import ProcessPoolExecutor
from datetime import datetime, timedelta
from pathlib import Path
from typing import TYPE_CHECKING

import adaptive
import cloudpickle
import pandas as pd
import pytest

from adaptive_scheduler import utils

if TYPE_CHECKING:
    from typing import Literal


def test_shuffle_list() -> None:
    """Test `utils.shuffle_list`."""
    list1 = [1, 2, 3, 4, 5]
    list2 = ["a", "b", "c", "d", "e"]
    shuffled1, shuffled2 = utils.shuffle_list(list1, list2, seed=42)
    assert shuffled1 == (4, 2, 3, 5, 1)
    assert shuffled2 == ("d", "b", "c", "e", "a")


def test_hash_anything() -> None:
    """Test `utils.hash_anything`."""
    assert utils.hash_anything("test_string") == "fe61a408ddf146f33219dfcfbfae6fe6"
    assert utils.hash_anything(12345) == "ab828429197deaaefe90ed273900e4a3"
    assert utils.hash_anything(["a", "b", "c"]) == "92c9643673b92ef3058d62fccca628e2"


def test_split() -> None:
    """Test `utils.split`."""
    seq = list(range(10))
    n_parts = 3
    result = list(utils.split(seq, n_parts))
    assert result == [(0, 1, 2, 3), (4, 5, 6, 7), (8, 9)]


def test_split_in_balancing_learners(
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test `utils.split_in_balancing_learners`."""
    n_parts = 2
    new_learners, new_fnames = utils.split_in_balancing_learners(
        learners,
        fnames,
        n_parts,
    )
    assert len(new_learners) == n_parts
    assert all(isinstance(lrn, adaptive.BalancingLearner) for lrn in new_learners)
    assert new_fnames == [[fnames[0]], [fnames[1]]]


def test_split_sequence_learner() -> None:
    """Test `utils.split_sequence_learner`."""
    big_learner = adaptive.SequenceLearner(lambda x: x, sequence=list(range(10)))
    n_learners = 3
    new_learners, new_fnames = utils.split_sequence_learner(big_learner, n_learners)
    assert len(new_learners) == n_learners
    assert all(isinstance(lrn, adaptive.SequenceLearner) for lrn in new_learners)


def test_split_sequence_in_sequence_learners() -> None:
    """Test `utils.split_sequence_in_sequence_learners`."""

    def function(x: int) -> int:
        return x

    sequence = list(range(10))
    n_learners = 3
    new_learners, new_fnames = utils.split_sequence_in_sequence_learners(
        function,
        sequence,
        n_learners,
    )
    assert len(new_learners) == n_learners
    assert all(isinstance(lrn, adaptive.SequenceLearner) for lrn in new_learners)
    assert len(new_fnames) == n_learners


def test_combine_sequence_learners() -> None:
    """Test `utils.combine_sequence_learners`."""
    learners = [
        adaptive.SequenceLearner(lambda x: x, sequence=list(range(5))),
        adaptive.SequenceLearner(lambda x: x, sequence=list(range(5, 10))),
    ]
    big_learner = utils.combine_sequence_learners(learners)
    assert isinstance(big_learner, adaptive.SequenceLearner)
    assert list(big_learner.sequence) == list(range(10))


def test_copy_from_sequence_learner() -> None:
    """Test `utils.copy_from_sequence_learner`."""
    learner1 = adaptive.SequenceLearner(lambda x: x, sequence=list(range(5)))
    learner2 = adaptive.SequenceLearner(lambda x: x, sequence=list(range(5)))
    for i, x in enumerate(learner1.sequence):
        learner1.tell((i, x), x * 2)
    utils.copy_from_sequence_learner(learner1, learner2)
    assert learner1.data == learner2.data


def test_get_npoints() -> None:
    """Test `utils._get_npoints`."""
    learner1 = adaptive.Learner1D(lambda x: x, bounds=(-10, 10))
    assert utils._get_npoints(learner1) == 0
    learner1.ask(10)
    learner1.tell(10, 42)
    assert utils._get_npoints(learner1) == 1


def test_progress() -> None:
    """Test `utils._progress`."""
    seq = list(range(10))
    with_progress_bar = False
    result = list(utils._progress(seq, with_progress_bar))
    assert result == seq


def test_combo_to_fname() -> None:
    """Test `utils.combo_to_fname`."""
    combo = {"x": 1, "y": 2, "z": 3}
    fname = utils.combo_to_fname(combo)
    assert str(fname) == "x_1__y_2__z_3.pickle"


def test_combo2fname() -> None:
    """Test `utils.combo2fname`."""
    combo = {"x": 1, "y": 2, "z": 3}
    fname = utils.combo2fname(combo)
    assert str(fname) == "x_1__y_2__z_3.pickle"


def test_add_constant_to_fname() -> None:
    """Test `utils.add_constant_to_fname`."""
    combo = {"x": 1, "y": 2, "z": 3}
    constant = {"a": 42}
    old_fname, new_fname = utils.add_constant_to_fname(combo, constant)
    assert str(old_fname) == "x_1__y_2__z_3.pickle"
    assert str(new_fname) == "a_42__x_1__y_2__z_3.pickle"


def test_maybe_round() -> None:
    """Test `utils.maybe_round`."""
    assert utils.maybe_round(3.14159265, 3) == 3.14
    assert utils.maybe_round(1 + 2j, 3) == 1 + 2j
    assert utils.maybe_round("test", 3) == "test"


def test_round_sigfigs() -> None:
    """Test `utils.round_sigfigs`."""
    assert utils.round_sigfigs(3.14159265, 3) == 3.14
    assert utils.round_sigfigs(123456789, 3) == 123000000


def test_remove_or_move_files(tmp_path: Path) -> None:
    """Test `utils._remove_or_move_files`."""
    test_file = tmp_path / "test.txt"
    test_file.write_text("test content")

    assert test_file.exists()

    utils._remove_or_move_files([str(test_file)], with_progress_bar=False)

    assert not test_file.exists()


def test_load_parallel(tmp_path: Path) -> None:
    """Test `utils.load_parallel`."""
    learners = [
        adaptive.Learner1D(lambda x: x, bounds=(-10, 10)),
        adaptive.SequenceLearner(lambda x: x, sequence=list(range(5))),
    ]
    fnames = [str(tmp_path / "learner1.pickle"), str(tmp_path / "learner2.pickle")]

    for learner, fname in zip(learners, fnames, strict=True):
        learner.save(fname)

    loaded_learners = [lrn.new() for lrn in learners]
    utils.load_parallel(loaded_learners, fnames, with_progress_bar=False)

    for learner, loaded_learner in zip(learners, loaded_learners, strict=True):
        assert learner.data == loaded_learner.data


def test_save_parallel(tmp_path: Path) -> None:
    """Test `utils.save_parallel`."""
    learners = [
        adaptive.Learner1D(lambda x: x, bounds=(-10, 10)),
        adaptive.SequenceLearner(lambda x: x, sequence=list(range(5))),
    ]
    fnames = [str(tmp_path / "learner1.pickle"), str(tmp_path / "learner2.pickle")]

    utils.save_parallel(learners, fnames, with_progress_bar=False)

    for fname in fnames:
        assert Path(fname).exists()


def test_connect_to_ipyparallel() -> None:
    """Test `utils.connect_to_ipyparallel`."""
    with pytest.raises(
        OSError,
        match="You have attempted to connect to an IPython Cluster",
    ):
        utils.connect_to_ipyparallel(n=1, profile="non_existent_profile", timeout=1)


def test_lru_cached_callable() -> None:
    """Test `utils.LRUCachedCallable`."""

    @utils.shared_memory_cache(cache_size=5)
    def cached_function(x: float) -> float:
        return x**2

    assert cached_function.cache_dict == {}
    assert cached_function(2) == 4
    assert (
        cached_function.cache_dict == {"{'x': 2}": 4}
        if sys.version_info >= (3, 9)
        else {"OrderedDict([('x', 2)])": 4}
    )

    assert cached_function(3) == 9
    assert (
        cached_function.cache_dict == {"{'x': 2}": 4, "{'x': 3}": 9}
        if sys.version_info >= (3, 9)
        else {"OrderedDict([('x', 2)]): 4, OrderedDict([('x', 3)]): 9"}
    )


def test_smart_goal() -> None:
    """Test `utils.smart_goal`."""
    learner = adaptive.Learner1D(lambda x: x, bounds=(-10, 10))

    goal_callable = utils.smart_goal(
        lambda lrn: lrn.npoints >= 10,
        [learner],
    )
    assert not goal_callable(learner)
    adaptive.runner.simple(learner, npoints_goal=5)
    assert not goal_callable(learner)
    adaptive.runner.simple(learner, npoints_goal=11)
    assert goal_callable(learner)

    goal_int = utils.smart_goal(10, [learner])
    assert goal_int(learner)

    goal_float = utils.smart_goal(0.1, [learner])
    adaptive.runner.simple(learner, loss_goal=0.09)
    assert goal_float(learner)

    goal_timedelta = utils.smart_goal(timedelta(milliseconds=50), [learner])
    assert not goal_timedelta(learner)
    time.sleep(0.1)
    assert goal_timedelta(learner)

    goal_datetime = utils.smart_goal(
        datetime.now() + timedelta(milliseconds=50),  # noqa: DTZ005
        [learner],
    )
    assert not goal_datetime(learner)
    time.sleep(0.1)
    assert goal_datetime(learner)

    goal_none = utils.smart_goal(None, [learner])
    assert not goal_none(learner)

    seq_learner = adaptive.SequenceLearner(lambda x: x, sequence=list(range(10)))
    goal_none = utils.smart_goal(None, [seq_learner])
    assert goal_none(seq_learner) == adaptive.SequenceLearner.done(seq_learner)


def test_time_goal() -> None:
    """Test `utils._TimeGoal`."""
    time_goal = utils._TimeGoal(timedelta(seconds=1))
    learner = adaptive.Learner1D(lambda x: x, bounds=(-10, 10))
    assert not time_goal(learner)
    time.sleep(1)
    assert time_goal(learner)


def test_fname_to_learner_fname() -> None:
    """Test `utils.fname_to_learner_fname`."""
    fname = "test.pickle"
    learner_fname = utils.fname_to_learner_fname(fname)
    assert learner_fname == Path(".learner.test.pickle")


def test_fname_to_learner(tmp_path: Path) -> None:
    """Test `utils.fname_to_learner`."""
    learner = adaptive.Learner1D(lambda x: x, bounds=(-10, 10))
    fname = tmp_path / "test.pickle"
    utils.cloudpickle_learners([learner], [fname], with_progress_bar=False)

    loaded_learner = utils.fname_to_learner(fname)
    assert isinstance(loaded_learner, adaptive.Learner1D)


def test_ensure_folder_exists(tmp_path: Path) -> None:
    """Test `utils._ensure_folder_exists`."""
    fnames = [tmp_path / "test1.pickle", tmp_path / "test2.pickle"]

    for fname in fnames:
        assert not Path(fname).exists()

    utils._ensure_folder_exists(fnames)

    for fname in fnames:
        assert Path(fname.parent).exists()


def test_cloudpickle_learners(tmp_path: Path) -> None:
    """Test `utils.cloudpickle_learners`."""
    learners = [
        adaptive.Learner1D(lambda x: x, bounds=(-10, 10)),
        adaptive.SequenceLearner(lambda x: x, sequence=list(range(5))),
    ]
    fnames = [tmp_path / "learner1.pickle", tmp_path / "learner2.pickle"]

    utils.cloudpickle_learners(learners, fnames, with_progress_bar=False)

    for fname in fnames:
        fname_learner = utils.fname_to_learner_fname(fname)
        assert fname_learner.exists()


@pytest.mark.parametrize("atomically", [False, True])
@pytest.mark.parametrize("fmt", ["pickle", "csv", "json"])
def test_save_dataframe(
    tmp_path: Path,
    atomically: bool,  # noqa: FBT001
    fmt: str,
) -> None:
    """Test `utils.save_dataframe`."""
    fmt = typing.cast("utils._DATAFRAME_FORMATS", fmt)
    learner = adaptive.Learner1D(lambda x: x, bounds=(-10, 10))
    fname = str(tmp_path / f"test.{fmt}")
    save_df = utils.save_dataframe(fname, atomically=atomically, format=fmt)

    learner.ask(10)
    learner.tell(10, 42)

    save_df(learner)

    assert utils.fname_to_dataframe(fname, format=fmt).exists()


@pytest.mark.parametrize("atomically", [False, True])
def test_load_dataframes(
    tmp_path: Path,
    atomically: bool,  # noqa: FBT001
) -> None:
    """Test `utils.load_dataframes`."""
    learner = adaptive.Learner1D(lambda x: x, bounds=(-10, 10))
    fname = str(tmp_path / "test.pickle")
    save_df = utils.save_dataframe(fname, atomically=atomically)

    learner.ask(10)
    learner.tell(10, 42)

    save_df(learner)

    df = utils.load_dataframes([fname], concat=False)[0]

    assert isinstance(df, pd.DataFrame)
    assert len(df) == 1


def test_expand_dict_columns() -> None:
    """Test `utils.expand_dict_columns`."""
    data = {"a": 1, "b": {"c": 2, "d": 3}}
    df = pd.DataFrame([data])

    expanded_df = utils.expand_dict_columns(df)

    assert "b" not in expanded_df.columns
    assert {"a", "b.c", "b.d"} == set(expanded_df.columns)


def test_remove_or_move_files_move(tmp_path: Path) -> None:
    """Test `utils._remove_or_move_files` with `move_to`."""
    test_file = tmp_path / "test.txt"
    test_file.write_text("test content")
    new_folder = tmp_path / "new_folder"
    new_folder.mkdir()

    assert test_file.exists()

    utils._remove_or_move_files(
        [test_file],
        with_progress_bar=False,
        move_to=new_folder,
    )

    assert not test_file.exists()
    assert (new_folder / "test.txt").exists()


def test_combo_to_fname_with_folder() -> None:
    """Test `utils.combo_to_fname` with `folder`."""
    combo = {"x": 1, "y": 2, "z": 3}
    folder = "test_folder"
    fname = utils.combo_to_fname(combo, folder=folder)
    assert fname == Path(f"{folder}/x_1__y_2__z_3.pickle")


def test_combo2fname_with_folder() -> None:
    """Test `utils.combo2fname` with `folder`."""
    combo = {"x": 1, "y": 2, "z": 3}
    folder = "test_folder"
    fname = utils.combo2fname(combo, folder=folder)
    assert fname == Path(f"{folder}/x_1__y_2__z_3.pickle")


def test_add_constant_to_fname_with_folder() -> None:
    """Test `utils.add_constant_to_fname` with `folder`."""
    combo = {"x": 1, "y": 2, "z": 3}
    constant = {"a": 42}
    folder = "test_folder"
    old_fname, new_fname = utils.add_constant_to_fname(
        combo,
        constant,
        folder=folder,
    )
    assert old_fname == Path(f"{folder}/x_1__y_2__z_3.pickle")
    assert new_fname == Path(f"{folder}/a_42__x_1__y_2__z_3.pickle")


def test_fname_to_dataframe_with_folder() -> None:
    """Test `utils.fname_to_dataframe` with `folder`."""
    fname = "test_folder/test.pickle"
    df_fname = utils.fname_to_dataframe(fname, "parquet")
    assert df_fname == Path("test_folder/dataframe.test.parquet")


def test_atomic_write(tmp_path: Path) -> None:
    """Ensure atomic_write works when operated in a basic manner."""
    path = tmp_path / "testfile"
    content = "this is some content"

    # Works correctly in basic mode, with no file existing
    with utils.atomic_write(path) as fp:
        fp.write(content)
    with path.open() as fp:
        assert content == fp.read()

    # Works correctly in basic mode, with the file existing
    assert path.exists()
    assert path.is_file()
    content = "this is some additional content"
    with utils.atomic_write(path) as fp:
        fp.write(content)
    with path.open() as fp:
        assert content == fp.read()

    # Works correctly when 'return_path' is used.
    content = "even more content"
    with (
        utils.atomic_write(path, return_path=True) as tmp_path2,
        tmp_path2.open("w") as fp,
    ):
        fp.write(content)
    with path.open() as fp:
        assert content == fp.read()


def test_atomic_write_no_write(tmp_path: Path) -> None:
    """Ensure atomic_write creates an empty file, if we do nothing.

    This gives 'atomic_write' the same semantics as 'open', when
    the file does not exist.
    """
    path = tmp_path / "testfile"

    assert not path.exists()  # Sanity check

    with utils.atomic_write(path) as _:
        pass
    assert path.stat().st_size == 0

    with utils.atomic_write(path, return_path=True) as _:
        pass
    assert path.stat().st_size == 0

    with utils.atomic_write(path) as fp:
        fp.write("content")
    assert path.stat().st_size > 0

    with utils.atomic_write(path) as _:
        pass
    assert path.stat().st_size == 0

    with utils.atomic_write(path) as fp:
        fp.write("content")
    assert path.stat().st_size > 0

    with utils.atomic_write(path, return_path=True) as _:
        pass
    assert path.stat().st_size == 0


def test_atomic_write_nested(tmp_path: Path) -> None:
    """Ensure nested calls to atomic_write on the same file work as expected."""
    path = tmp_path / "testfile"
    with (
        utils.atomic_write(path, mode="w") as fp,
        utils.atomic_write(path, mode="w") as fp2,
    ):
        fp.write("one")
        fp2.write("two")
    # Outer call wins
    with path.open() as fp:
        assert fp.read() == "one"


@pytest.mark.parametrize("atomically", [False, True])
def test_load_dataframes_with_folder(
    tmp_path: Path,
    atomically: bool,  # noqa: FBT001
) -> None:
    """Test `utils.load_dataframes` with `folder`."""
    folder = tmp_path / "test_folder"
    folder.mkdir()
    learner = adaptive.Learner1D(lambda x: x, bounds=(-10, 10))
    fname = str(folder / "test.pickle")
    save_df = utils.save_dataframe(fname, atomically=atomically)

    learner.ask(10)
    learner.tell(10, 42)

    save_df(learner)

    df = utils.load_dataframes([fname], concat=False)[0]

    assert isinstance(df, pd.DataFrame)
    assert len(df) == 1


def test_at_least_adaptive_version() -> None:
    """Test `utils._at_least_adaptive_version`."""
    assert utils._at_least_adaptive_version("0.0.0", raises=False)
    assert utils._at_least_adaptive_version("0.0.0")
    assert not utils._at_least_adaptive_version("100000.0.0", raises=False)
    with pytest.raises(RuntimeError, match="requires adaptive version"):
        assert utils._at_least_adaptive_version("100000.0.0", raises=True)
    utils._at_least_adaptive_version(adaptive.__version__)


def _is_key_in_global_cache(key: str | bytes) -> bool:
    return key in utils._GLOBAL_CACHE


@pytest.mark.parametrize("mode", ["file", "memory", "random_id"])
def test_executor_with_wrapped_function_that_is_loaded_with_cloudpickle(
    *,
    mode: Literal["memory", "random_id", "file"],
) -> None:
    """Test executor with WrappedFunction that is loaded with cloudpickle."""
    if mode == "random_id" and platform.system() == "Darwin":
        pytest.skip("Not possible on MacOS")

    # Define a simple test function
    def square(x: int) -> int:
        return x * x

    # Serialize the function using cloudpickle
    serialized_function = cloudpickle.dumps(square)

    # Remove the function from the current scope
    del square

    # Load the serialized function using cloudpickle
    loaded_function = cloudpickle.loads(serialized_function)

    # Wrap the loaded function using WrappedFunction
    wrapped_function = utils.WrappedFunction(
        loaded_function,
        mode=mode,
    )

    # Run the wrapped function using ProcessPoolExecutor
    ex = ProcessPoolExecutor()

    # Check if the global cache contains the key in the cache
    # Note that behaviour on Linux and MacOS is different due to 'fork' vs 'spawn'
    fut_is_key = ex.submit(_is_key_in_global_cache, wrapped_function._cache_key)
    is_key = fut_is_key.result()
    if platform.system() == "Darwin":
        assert not is_key
    elif platform.system() == "Linux":
        assert is_key

    # Note that passing the loaded_function directly to ex.submit will not work!
    fut = ex.submit(wrapped_function, 4)
    result = fut.result()

    assert result == 16, f"Expected 16, but got {result}"

    # Check if the global cache contains the key in the executor
    fut_is_key = ex.submit(_is_key_in_global_cache, wrapped_function._cache_key)
    is_key = fut_is_key.result()
    assert is_key


def test_datetime_now() -> None:
    """Test `utils.datetime_now`."""
    # Test that the functions are compatible and can roundtrip a datetime string
    now = utils._now()
    now_roundtripped = utils._from_datetime(now)

    assert now == now_roundtripped.isoformat()
    assert utils._time_between(now, now) == 0



from __future__ import annotations

import asyncio
from pathlib import Path
from typing import TYPE_CHECKING

from adaptive_scheduler.utils import (
    _remove_or_move_files,
    sleep_unless_task_is_done,
)

from .base_manager import BaseManager
from .common import _maybe_path, log

if TYPE_CHECKING:
    from collections.abc import Callable

    from adaptive_scheduler.scheduler import BaseScheduler

    from .database_manager import DatabaseManager


def logs_with_string_or_condition(
    error: str | Callable[[list[str]], bool],
    database_manager: DatabaseManager,
) -> list[tuple[str, list[str]]]:
    """Get jobs that have `string` (or apply a callable) inside their log-file.

    Either use `string` or `error`.

    Parameters
    ----------
    error
        String that is searched for or callable that is applied
        to the log text. Must take a single argument, a list of
        strings, and return True if the job has to be killed, or
        False if not.
    database_manager
        A `DatabaseManager` instance.

    Returns
    -------
    has_string
        A list ``(job_name, fnames)``, which have the string inside their log-file.

    """
    if isinstance(error, str):
        has_error = lambda lines: error in "".join(lines)  # noqa: E731
    elif callable(error):
        has_error = error
    else:
        msg = "`error` can only be a `str` or `callable`."
        raise TypeError(msg)

    def file_has_error(fname: Path) -> bool:
        if not fname.exists():
            return False
        with fname.open(encoding="utf-8") as f:
            lines = f.readlines()
        return has_error(lines)

    have_error = []
    for entry in database_manager.as_dicts():
        fnames = entry["output_logs"]
        if entry["job_id"] is not None and any(file_has_error(Path(f)) for f in fnames):
            all_fnames = [*fnames, entry["log_fname"]]
            have_error.append((entry["job_name"], all_fnames))
    return have_error


class KillManager(BaseManager):
    """Kill manager.

    Automatically cancel jobs that contain an error (or other condition)
    in the log files.

    Parameters
    ----------
    scheduler
        A scheduler instance from `adaptive_scheduler.scheduler`.
    database_manager
        A `DatabaseManager` instance.
    error
        If ``error`` is a string and is found in the log files, the job will
        be cancelled and restarted. If it is a callable, it is applied
        to the log text. Must take a single argument, a list of
        strings, and return True if the job has to be killed, or
        False if not.
    interval
        Time in seconds between checking for the condition.
    max_cancel_tries
        Try maximum `max_cancel_tries` times to cancel a job.
    move_to
        If a job is cancelled the log is either removed (if ``move_to=None``)
        or moved to a folder (e.g. if ``move_to='old_logs'``).

    """

    def __init__(
        self,
        scheduler: BaseScheduler,
        database_manager: DatabaseManager,
        *,
        error: str | Callable[[list[str]], bool] = "srun: error:",
        interval: float = 600,
        max_cancel_tries: int = 5,
        move_to: str | Path | None = None,
    ) -> None:
        super().__init__()
        self.scheduler = scheduler
        self.database_manager = database_manager
        self.error = error
        self.interval = interval
        self.max_cancel_tries = max_cancel_tries
        self.move_to = _maybe_path(move_to)

        self.cancelled: list[str] = []
        self.deleted: list[str] = []

    async def _manage(self) -> None:
        while True:
            try:
                self.database_manager.update()
                failed_jobs = logs_with_string_or_condition(
                    self.error,
                    self.database_manager,
                )

                to_cancel: list[str] = []
                to_delete: list[str] = []
                for job_name, fnames in failed_jobs:
                    to_cancel.append(job_name)
                    to_delete.extend(fnames)

                self.scheduler.cancel(
                    to_cancel,
                    with_progress_bar=False,
                    max_tries=self.max_cancel_tries,
                )
                _remove_or_move_files(
                    to_delete,
                    with_progress_bar=False,
                    move_to=self.move_to,
                )
                self.cancelled.extend(to_cancel)
                self.deleted.extend(to_delete)

                if await sleep_unless_task_is_done(
                    self.database_manager.task,  # type: ignore[arg-type]
                    self.interval,
                ):  # if true, we are done
                    return
            except asyncio.CancelledError:  # noqa: PERF203
                log.info("task was cancelled because of a CancelledError")
                raise
            except Exception as e:  # noqa: BLE001
                log.exception("got exception in kill manager", exception=str(e))
                if await sleep_unless_task_is_done(
                    self.database_manager.task,  # type: ignore[arg-type]
                    self.interval,
                ):  # if true, we are done
                    return



"""Adaptive Scheduler."""

from adaptive_scheduler import client_support, scheduler, server_support, utils
from adaptive_scheduler._executor import SlurmExecutor, SlurmTask
from adaptive_scheduler._version import __version__
from adaptive_scheduler.scheduler import PBS, SLURM
from adaptive_scheduler.server_support import (
    MultiRunManager,
    RunManager,
    slurm_run,
    start_one_by_one,
)

__all__ = [
    "PBS",
    "SLURM",
    "MultiRunManager",
    "RunManager",
    "SlurmExecutor",
    "SlurmTask",
    "__version__",
    "client_support",
    "scheduler",
    "server_support",
    "slurm_run",
    "start_one_by_one",
    "utils",
]



"""Test the file creation progress tracking utilities."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest
from rich.progress import Progress, TaskID

from adaptive_scheduler.utils import (
    _remove_completed_paths,
    _track_file_creation_progress,
    _update_progress_for_paths,
)

if TYPE_CHECKING:
    from pathlib import Path


def test_update_progress_for_paths(tmp_path: Path) -> None:
    """Test the update progress for paths function."""
    # Create test files
    create_test_files(tmp_path, ["file1", "file3", "file4"])

    paths_dict: dict[str, set[Path | tuple[Path, ...]]] = {
        "category1": {tmp_path / "file1", tmp_path / "file2"},
        "category2": {(tmp_path / "file3", tmp_path / "file4")},
    }

    progress_mock = Mock(spec=Progress)
    task_id_mock = Mock(spec=TaskID)
    task_ids = {"category1": task_id_mock, "category2": task_id_mock}

    processed = _update_progress_for_paths(
        paths_dict,
        progress_mock,
        task_id_mock,
        task_ids,
    )

    assert processed == 2  # Only "file1" and the tuple ("file3", "file4") exist
    assert len(paths_dict["category1"]) == 1  # "file2" does not exist and should remain
    assert len(paths_dict["category2"]) == 0  # Tuple paths exist and should be removed
    progress_mock.update.assert_called()


def create_test_files(tmp_path: Path, file_names: list[str]) -> None:
    """Create test files in the given directory."""
    for name in file_names:
        (tmp_path / name).touch()


def test_remove_completed_paths(tmp_path: Path) -> None:
    """Test the remove completed paths function."""
    # Create test files
    existing_files = ["file1", "file3", "file4"]
    create_test_files(tmp_path, existing_files)

    paths_dict: dict[str, set[Path | tuple[Path, ...]]] = {
        "category1": {tmp_path / "file1", tmp_path / "file2"},
        "category2": {(tmp_path / "file3", tmp_path / "file4")},
    }

    n_completed = _remove_completed_paths(paths_dict)

    assert n_completed == {"category1": 1, "category2": 1}
    assert paths_dict == {"category1": {tmp_path / "file2"}, "category2": set()}


@pytest.mark.asyncio
async def test_track_file_creation_progress(tmp_path: Path) -> None:
    """Test the track file creation progress function."""
    # Create test files
    create_test_files(tmp_path, ["file1"])

    paths_dict: dict[str, set[Path | tuple[Path, ...]]] = {
        "category1": {tmp_path / "file1", tmp_path / "file2"},
        "category2": {(tmp_path / "file3", tmp_path / "file4")},
    }

    progress = Progress(auto_refresh=False)
    task = asyncio.create_task(
        _track_file_creation_progress(paths_dict, progress, interval=1e-3),
    )

    # Allow some time for the task to process
    await asyncio.sleep(0.1)

    progress.stop()
    assert "Total" in progress._tasks[0].description
    assert progress._tasks[0].total == 3
    assert progress._tasks[0].completed == 1

    assert "category1" in progress._tasks[1].description
    assert progress._tasks[1].total == 2
    assert progress._tasks[1].completed == 1

    assert "category2" in progress._tasks[2].description
    assert progress._tasks[2].total == 1
    assert progress._tasks[2].completed == 0

    # Create one of the files of category2, should still not be completed
    create_test_files(tmp_path, ["file3"])
    await asyncio.sleep(0.1)

    assert "category2" in progress._tasks[2].description
    assert progress._tasks[2].total == 1
    assert progress._tasks[2].completed == 0
    assert paths_dict["category2"] == {(tmp_path / "file4",)}

    # Create the other file of category2, should now be completed
    create_test_files(tmp_path, ["file4"])
    await asyncio.sleep(0.05)
    assert "category2" in progress._tasks[2].description
    assert progress._tasks[2].total == 1
    assert progress._tasks[2].completed == 1
    assert paths_dict["category2"] == set()

    # Create the other file of category1, should now be completed
    create_test_files(tmp_path, ["file2"])
    await asyncio.sleep(0.05)
    assert "category1" in progress._tasks[1].description
    assert progress._tasks[1].total == 2
    assert progress._tasks[1].completed == 2

    # Check the total progress
    assert "Total" in progress._tasks[0].description
    assert progress._tasks[0].total == 3
    assert progress._tasks[0].completed == 3

    # Stop the progress and the task
    progress.stop()
    task.cancel()



"""Test the MultiRunManager class."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING
from unittest.mock import patch

import ipywidgets as ipw
import pytest

from adaptive_scheduler._server_support.multi_run_manager import MultiRunManager
from adaptive_scheduler._server_support.run_manager import RunManager

if TYPE_CHECKING:
    from pathlib import Path

    import adaptive

    from .helpers import MockScheduler


@pytest.fixture
def mock_run_manager(
    mock_scheduler: MockScheduler,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> RunManager:
    """Create a mock RunManager for testing."""
    return RunManager(mock_scheduler, learners[:1], fnames[:1], job_name="test-rm")


def test_multi_run_manager_init() -> None:
    """Test the initialization of MultiRunManager."""
    mrm = MultiRunManager()
    assert isinstance(mrm, MultiRunManager)
    assert mrm.run_managers == {}


def test_multi_run_manager_add_run_manager(mock_run_manager: RunManager) -> None:
    """Test adding a RunManager to MultiRunManager."""
    mrm = MultiRunManager()
    mrm.add_run_manager(mock_run_manager)
    assert len(mrm.run_managers) == 1
    assert "test-rm" in mrm.run_managers
    assert mrm.run_managers["test-rm"] == mock_run_manager


def test_multi_run_manager_add_duplicate_run_manager(mock_run_manager: RunManager) -> None:
    """Test adding a duplicate RunManager to MultiRunManager."""
    mrm = MultiRunManager()
    mrm.add_run_manager(mock_run_manager)
    with pytest.raises(ValueError, match="A RunManager with the name 'test-rm' already exists."):
        mrm.add_run_manager(mock_run_manager)


@pytest.mark.asyncio
async def test_multi_run_manager_add_run_manager_with_start(mock_run_manager: RunManager) -> None:
    """Test adding a RunManager to MultiRunManager with start=True."""
    mrm = MultiRunManager()
    mrm.add_run_manager(mock_run_manager, start=True)
    await asyncio.sleep(0.1)
    assert mock_run_manager.status() == "running"
    mock_run_manager.cancel()


def test_multi_run_manager_add_run_manager_with_invalid_wait_for(
    mock_run_manager: RunManager,
) -> None:
    """Test adding a RunManager with an invalid wait_for parameter."""
    mrm = MultiRunManager()
    with pytest.raises(KeyError, match="No RunManager with the name 'non-existent' exists."):
        mrm.add_run_manager(mock_run_manager, start=True, wait_for="non-existent")


def test_multi_run_manager_add_run_manager_with_wait_for_without_start(
    mock_run_manager: RunManager,
) -> None:
    """Test adding a RunManager with wait_for but without start=True."""
    mrm = MultiRunManager()
    mrm.add_run_manager(
        RunManager(
            mock_run_manager.scheduler,
            mock_run_manager.learners,
            mock_run_manager.fnames,
            job_name="rm1",
        ),
    )
    with pytest.raises(ValueError, match="`start` must be True if `wait_for` is used."):
        mrm.add_run_manager(mock_run_manager, wait_for="rm1")


def test_multi_run_manager_remove_run_manager(mock_run_manager: RunManager) -> None:
    """Test removing a RunManager from MultiRunManager."""
    mrm = MultiRunManager()
    mrm.add_run_manager(mock_run_manager)
    mrm.remove_run_manager("test-rm")
    assert len(mrm.run_managers) == 0


def test_multi_run_manager_remove_non_existent_run_manager() -> None:
    """Test removing a non-existent RunManager from MultiRunManager."""
    mrm = MultiRunManager()
    with pytest.raises(KeyError, match="No RunManager with the name 'non-existent' exists."):
        mrm.remove_run_manager("non-existent")


@pytest.mark.asyncio
async def test_multi_run_manager_start_all(mock_run_manager: RunManager) -> None:
    """Test starting all RunManagers in MultiRunManager."""
    mrm = MultiRunManager()
    mrm.add_run_manager(mock_run_manager)
    mrm.start_all()
    await asyncio.sleep(0.1)
    assert mock_run_manager.status() == "running"
    mock_run_manager.cancel()


@pytest.mark.asyncio
async def test_multi_run_manager_cancel_all(mock_run_manager: RunManager) -> None:
    """Test cancelling all RunManagers in MultiRunManager."""
    mrm = MultiRunManager()
    mrm.add_run_manager(mock_run_manager)
    mrm.start_all()
    await asyncio.sleep(0.1)
    mrm.cancel_all()
    await asyncio.sleep(0.1)
    assert mock_run_manager.status() == "cancelled"


def test_multi_run_manager_create_widget(mock_run_manager: RunManager) -> None:
    """Test creating the widget for MultiRunManager."""
    mrm = MultiRunManager()
    mrm.add_run_manager(mock_run_manager)
    vbox = mrm._create_widget()
    assert isinstance(vbox, ipw.VBox)
    tab = vbox.children[1]
    assert isinstance(tab, ipw.Tab)
    assert len(tab.children) == 1
    assert tab.get_title(0) == "test-rm"


def test_multi_run_manager_update_widget(mock_run_manager: RunManager) -> None:
    """Test updating the widget for MultiRunManager."""
    mrm = MultiRunManager()
    mrm.add_run_manager(mock_run_manager)
    mrm._widget = mrm._create_widget()
    new_rm = RunManager(
        mock_run_manager.scheduler,
        mock_run_manager.learners,
        mock_run_manager.fnames,
        job_name="new-rm",
    )
    mrm.add_run_manager(new_rm)
    assert len(mrm._widget.children) == 2
    assert mrm._widget.children[1].get_title(1) == "new-rm"


def test_multi_run_manager_info(mock_run_manager: RunManager) -> None:
    """Test the info method of MultiRunManager."""
    mrm = MultiRunManager()
    mrm.add_run_manager(mock_run_manager)
    vbox = mrm.info()
    assert isinstance(vbox, ipw.VBox)
    tab = vbox.children[1]
    assert isinstance(tab, ipw.Tab)
    assert len(tab.children) == 1


def test_multi_run_manager_repr_html(mock_run_manager: RunManager) -> None:
    """Test the _repr_html_ method of MultiRunManager."""
    mrm = MultiRunManager()
    mrm.add_run_manager(mock_run_manager)
    with patch("IPython.display.display") as mocked_display:
        mrm._repr_html_()
        assert mocked_display.called



"""Tests for DatabaseManager."""

from __future__ import annotations

import asyncio
from pathlib import Path

import adaptive
import pytest
import zmq

from adaptive_scheduler._server_support.database_manager import (
    DatabaseManager,
    SimpleDatabase,
    _DBEntry,
    _ensure_str,
)
from adaptive_scheduler.utils import fname_to_learner_fname, smart_goal

from .helpers import send_message


def test_simple_database_init_and_save(tmp_path: Path) -> None:
    """Test initializing and saving a SimpleDatabase instance."""
    db_fname = tmp_path / "test_db.json"
    db = SimpleDatabase(db_fname)
    assert db.all() == []
    db._save()
    assert db_fname.exists()


def test_simple_database_insert_multiple(tmp_path: Path) -> None:
    """Test inserting multiple entries into the database."""
    db_fname = tmp_path / "test_db.json"
    db = SimpleDatabase(db_fname)
    entries = [
        _DBEntry(fname="file1.txt"),
        _DBEntry(fname="file2.txt"),
        _DBEntry(fname="file3.txt"),
    ]
    db.insert_multiple(entries)
    assert len(db.all()) == 3


def test_simple_database_update(tmp_path: Path) -> None:
    """Test updating entries in the database."""
    db_fname = tmp_path / "test_db.json"
    db = SimpleDatabase(db_fname)
    entries = [
        _DBEntry(fname="file1.txt"),
        _DBEntry(fname="file2.txt"),
        _DBEntry(fname="file3.txt"),
    ]
    db.insert_multiple(entries)
    db.update({"is_done": True}, indices=[1])
    assert db.all()[1].is_done is True


def test_simple_database_count(tmp_path: Path) -> None:
    """Test counting entries in the database."""
    db_fname = tmp_path / "test_db.json"
    db = SimpleDatabase(db_fname)
    entries = [
        _DBEntry(fname="file1.txt", is_done=True),
        _DBEntry(fname="file2.txt"),
        _DBEntry(fname="file3.txt", is_done=True),
    ]
    db.insert_multiple(entries)
    count_done = db.count(lambda entry: entry.is_done)
    assert count_done == 2


def test_simple_database_get_and_contains(tmp_path: Path) -> None:
    """Test getting and checking for entries in the database."""
    db_fname = tmp_path / "test_db.json"
    db = SimpleDatabase(db_fname)
    entries = [
        _DBEntry(fname="file1.txt"),
        _DBEntry(fname="file2.txt"),
        _DBEntry(fname="file3.txt"),
    ]
    db.insert_multiple(entries)
    entry = db.get(lambda entry: entry.fname == "file2.txt")
    assert entry is not None
    assert entry.fname == "file2.txt"
    assert not db.contains(lambda entry: entry.fname == "file4.txt")


def test_simple_database_get_all(tmp_path: Path) -> None:
    """Test getting all entries in the database."""
    db_fname = tmp_path / "test_db.json"
    db = SimpleDatabase(db_fname)
    entries = [
        _DBEntry(fname="file1.txt", is_done=True),
        _DBEntry(fname="file2.txt"),
        _DBEntry(fname="file3.txt", is_done=True),
    ]
    db.insert_multiple(entries)
    done_entries = db.get_all(lambda entry: entry.is_done)
    assert len(done_entries) == 2
    assert done_entries[0][1].fname == "file1.txt"
    assert done_entries[1][1].fname == "file3.txt"


@pytest.mark.asyncio
async def test_database_manager_start_and_cancel(db_manager: DatabaseManager) -> None:
    """Test starting and canceling the DatabaseManager."""
    db_manager.start()
    await asyncio.sleep(0.1)  # Give it some time to start
    assert db_manager.is_started
    with pytest.raises(Exception, match="already started"):
        db_manager.start()
    result = db_manager.cancel()
    assert result is not None
    with pytest.raises(asyncio.InvalidStateError):  # noqa: PT012
        assert db_manager.task is not None
        assert db_manager.task.result()


def test_database_manager_n_done(db_manager: DatabaseManager) -> None:
    """Test the number of done jobs."""
    assert db_manager.n_done() == 0


def test_smart_goal(learners: list) -> None:
    """Test empty learners didn't reach the goal."""
    goal = smart_goal(100, learners)
    assert not goal(learners[0])
    assert not goal(learners[1])
    goal = smart_goal(0, learners)
    assert goal(learners[0])


def test_database_manager_create_empty_db(db_manager: DatabaseManager) -> None:
    """Test creating an empty database."""
    db_manager.create_empty_db()
    assert db_manager._db is not None
    assert Path(db_manager.db_fname).exists()
    n_learners = 2
    assert len(db_manager._db.all()) == n_learners


def test_database_manager_as_dicts(
    db_manager: DatabaseManager,
    fnames: list[Path] | list[str],
) -> None:
    """Test getting the database as a list of dictionaries."""
    db_manager.create_empty_db()
    assert db_manager.as_dicts() == [
        {
            "fname": _ensure_str(fnames[0]),
            "is_pending": False,
            "is_done": False,
            "job_id": None,
            "job_name": None,
            "log_fname": None,
            "output_logs": [],
            "start_time": None,
            "depends_on": [],
        },
        {
            "fname": _ensure_str(fnames[1]),
            "is_pending": False,
            "is_done": False,
            "job_id": None,
            "job_name": None,
            "log_fname": None,
            "output_logs": [],
            "start_time": None,
            "depends_on": [],
        },
    ]


@pytest.mark.asyncio
async def test_database_manager_dispatch_start_stop(
    db_manager: DatabaseManager,
    learners: list[adaptive.Learner1D]
    | list[adaptive.BalancingLearner]
    | list[adaptive.SequenceLearner],
    fnames: list[str] | list[Path],
) -> None:
    """Test starting and stopping jobs using the dispatch method."""
    db_manager.learners, db_manager.fnames = learners, fnames
    db_manager.create_empty_db()
    index, _fname = db_manager._choose_fname()
    db_manager._confirm_submitted(index, "test_job")
    assert index == 0  # The first learner is chosen
    start_request = ("start", "1000", "log_1000.txt", "test_job")
    fname = db_manager._dispatch(start_request)  # type: ignore[arg-type]
    assert fname == _fname
    assert fname in _ensure_str(db_manager.fnames)
    if isinstance(learners[0], adaptive.BalancingLearner):
        assert isinstance(fname, list)
        assert isinstance(fname[0], str)
    else:
        assert isinstance(fname, str)

    stop_request = ("stop", fname)
    db_manager._dispatch(stop_request)
    assert db_manager._db is not None
    entry = db_manager._db.get(lambda entry: entry.fname == fname)
    assert entry is not None
    assert entry.job_id is None
    assert entry.is_done is True


@pytest.mark.asyncio
async def test_database_manager_start_and_update(
    socket: zmq.asyncio.Socket,
    db_manager: DatabaseManager,
    fnames: list[str] | list[Path],
) -> None:
    """Test starting and updating jobs."""
    db_manager.create_empty_db()
    db_manager.start()
    await asyncio.sleep(0.1)  # Give it some time to start

    job_id, log_fname, job_name = "1000", "log.log", "job_name"

    # Choose fname for "job_name" and tell the DB the job is launched
    index, _fname = db_manager._choose_fname()
    db_manager._confirm_submitted(index, job_name)

    # Send a start message to the DatabaseManager
    start_message = ("start", job_id, log_fname, job_name)
    fname = await send_message(socket, start_message)
    assert fname == _fname

    # Check if the correct fname is returned
    assert fname == _ensure_str(fnames[0]), fname

    # Check that the database is updated correctly
    assert db_manager._db is not None
    entry = db_manager._db.get(lambda entry: entry.fname == fname)
    assert entry is not None
    assert entry.job_id == job_id
    assert entry.log_fname == log_fname
    assert entry.job_name == job_name

    # Say that the job is still running
    queue = {"1000": {"job_id": job_id, "job_name": job_name}}
    db_manager.update(queue)

    # Check that the database is the same
    entry = db_manager._db.get(lambda entry: entry.fname == fname)
    assert entry is not None
    assert entry.job_id == job_id
    assert entry.log_fname == log_fname
    assert entry.job_name == job_name

    # Say that the job is died
    queue = {}
    db_manager.update(queue)

    # Check that the database is updated correctly
    entry = db_manager._db.get(lambda entry: entry.fname == fname)
    assert entry is not None
    assert entry.job_id is None


@pytest.mark.asyncio
async def test_database_manager_start_stop(
    socket: zmq.asyncio.Socket,
    db_manager: DatabaseManager,
    fnames: list[str] | list[Path],
) -> None:
    """Test starting and stopping jobs."""
    db_manager.create_empty_db()
    db_manager.start()
    await asyncio.sleep(0.1)  # Give it some time to start
    assert db_manager.task is not None

    job_id, log_fname, job_name = "1000", "log.log", "job_name"

    # Choose fname for "job_name" and tell the DB the job is launched
    index, _ = db_manager._choose_fname()
    db_manager._confirm_submitted(index, job_name)

    # Send a start message to the DatabaseManager
    start_message = ("start", job_id, log_fname, job_name)
    fname = await send_message(socket, start_message)
    # Try starting again:
    exception = await send_message(socket, start_message)
    assert "The job_id 1000 already exists in the database and runs" in str(exception)
    assert fname == _ensure_str(fnames[0]), fname

    assert db_manager._db is not None
    # Check that the database is updated correctly
    entry = db_manager._db.get(lambda entry: entry.fname == fname)
    assert entry is not None
    assert entry.job_id == job_id
    assert entry.log_fname == log_fname
    assert entry.job_name == job_name

    # Check that task is still running
    assert db_manager.task is not None
    assert not db_manager.task.done()

    # Send a stop message to the DatabaseManager
    stop_message = ("stop", fname)
    reply = await send_message(socket, stop_message)
    assert reply is None

    assert db_manager._db is not None
    entry = db_manager._db.get(lambda entry: entry.fname == _ensure_str(fnames[0]))
    assert entry is not None
    assert entry.job_id is None

    # Start and stop the learner2
    index2, _ = db_manager._choose_fname()
    db_manager._confirm_submitted(index2, job_name)
    fname = await send_message(socket, start_message)
    assert fname == _ensure_str(fnames[1])

    # Send a stop message to the DatabaseManager
    stop_message = ("stop", fname)
    reply = await send_message(socket, stop_message)
    assert reply is None

    with pytest.raises(zmq.error.Again, match="Resource temporarily unavailable"):
        await send_message(socket, start_message)


@pytest.mark.asyncio
async def test_database_manager_stop_request_and_requests(
    socket: zmq.asyncio.Socket,
    db_manager: DatabaseManager,
    fnames: list[str] | list[Path],
) -> None:
    """Test stopping jobs using stop_request and stop_requests methods."""
    db_manager.create_empty_db()
    db_manager.start()
    await asyncio.sleep(0.1)  # Give it some time to start
    assert db_manager.task is not None
    assert db_manager._db is not None

    # Start a job for learner1
    job_id1, log_fname1, job_name1 = "1000", "log1.log", "job_name1"
    start_message1 = ("start", job_id1, log_fname1, job_name1)

    _index1, _fname1 = db_manager._choose_fname()
    db_manager._confirm_submitted(_index1, job_name1)

    fname1 = await send_message(socket, start_message1)
    assert fname1 == _fname1
    assert fname1 == _ensure_str(fnames[0]), fname1
    e = db_manager._db.get(lambda entry: entry.fname == fname1)
    assert isinstance(e, _DBEntry)
    assert e.job_id == job_id1

    # Start a job for learner2
    job_id2, log_fname2, job_name2 = "1001", "log2.log", "job_name2"
    start_message2 = ("start", job_id2, log_fname2, job_name2)
    _index2, _fname2 = db_manager._choose_fname()
    db_manager._confirm_submitted(_index2, job_name2)
    fname2 = await send_message(socket, start_message2)
    assert fname2 == _fname2
    assert fname2 == _ensure_str(fnames[1]), fname2
    e = db_manager._db.get(lambda entry: entry.fname == fname2)
    assert isinstance(e, _DBEntry)
    assert e.job_id == job_id2

    # Stop the job for learner1 using _stop_request
    db_manager._stop_request(fname1)
    entry = db_manager._db.get(lambda entry: entry.fname == fname1)
    assert entry is not None
    assert entry.job_id is None, (fname1, fname2)
    assert entry.is_done is True
    assert entry.job_name is None

    # Stop the job for learner2 using _stop_requests
    db_manager._stop_requests([fname2])

    entry = db_manager._db.get(lambda entry: entry.fname == fname2)
    assert entry is not None
    assert entry.job_id is None, (fname1, fname2)
    assert entry.is_done is True
    assert entry.job_name is None


def test_job_failure_after_start_request(db_manager: DatabaseManager) -> None:
    """Ensure jobs that fail after they send a start request are updated in the DB."""
    db_manager.create_empty_db()
    assert not db_manager.failed

    index, _ = db_manager._choose_fname()
    db_manager._confirm_submitted(index, "job1")
    # job is launched and appears in queue
    db_manager.update({"1": {"job_name": "job1"}})
    # job reports back that it started
    db_manager._start_request("1", "log", "job1")
    # job dies, disappearing from queue
    db_manager.update({})
    assert db_manager.failed


def test_job_failure_before_start_request(db_manager: DatabaseManager) -> None:
    """Ensure jobs that fail before they send a start request are updated in the DB.

    This is a regression test against
    https://github.com/basnijholt/adaptive-scheduler/issues/216
    """
    db_manager.create_empty_db()
    assert not db_manager.failed

    index, _ = db_manager._choose_fname()
    # Job is launched
    db_manager._confirm_submitted(index, "job1")
    # Job appears in queue, and one of our "update" calls sees it
    db_manager.update({"1": {"job_name": "job1"}})
    # Job dies, disappearing from queue
    db_manager.update({})
    assert db_manager.failed


def test_job_failure_before_start_request_slow_update(
    db_manager: DatabaseManager,
) -> None:
    """Ensure jobs that fail before they send a start request are updated in the DB, even if db manager updates slowly.

    This is a regression test against
    https://github.com/basnijholt/adaptive-scheduler/issues/216
    """
    db_manager.create_empty_db()
    assert not db_manager.failed

    index, _ = db_manager._choose_fname()
    # Job is launched...
    db_manager._confirm_submitted(index, "job1")
    # But dies before the next time our DB updates.
    db_manager.update({})
    assert db_manager.failed


def test_job_failure_before_start_request_interleaved(
    db_manager: DatabaseManager,
) -> None:
    """Ensure jobs that fail before they send a start request are updated in the DB, even if db manager updates slowly.

    This is a regression test against
    https://github.com/basnijholt/adaptive-scheduler/issues/216
    """
    db_manager.create_empty_db()
    assert not db_manager.failed

    # fname is chosen...
    index, _ = db_manager._choose_fname()
    # Database is updated by another task efore launch is confirmed;
    # this does not result in the job being marked as failed.
    db_manager.update({})
    assert not db_manager.failed
    # Launch is later confirmed...
    db_manager._confirm_submitted(index, "job1")
    # But job is still not in queue...
    db_manager.update({})
    # which means that it will be marked as failed
    assert db_manager.failed


def test_job_failure_before_start_request_interleaved2(
    db_manager: DatabaseManager,
) -> None:
    """Ensure jobs that fail before they send a start request are updated in the DB, even if db manager updates slowly.

    This is a regression test against
    https://github.com/basnijholt/adaptive-scheduler/issues/216
    """
    db_manager.create_empty_db()
    assert not db_manager.failed

    # fname is chosen...
    index, _ = db_manager._choose_fname()
    # Database is updated after job is in the queue, but before
    # launch is confirmed by the job manager
    db_manager.update({"1": {"job_name": "job1"}})
    assert not db_manager.failed
    # launch is confirmed by job manager
    db_manager._confirm_submitted(index, "job1")
    # job disappears from queue...
    db_manager.update({})
    # which means that it will be marked as failed
    assert db_manager.failed


@pytest.mark.parametrize(
    ("input_list", "expected_output"),
    [
        # Test with a list of strings
        (["path1", "path2", "path3", "path4"], ["path1", "path2", "path3", "path4"]),
        # Test with a list of Path objects
        (
            [Path("path1"), Path("path2"), Path("path3"), Path("path4")],
            ["path1", "path2", "path3", "path4"],
        ),
        # Test with a list of lists of strings
        (
            [["path1", "path2"], ["path3", "path4"]],
            [["path1", "path2"], ["path3", "path4"]],
        ),
        # Test with a list of lists of Path objects
        (
            [[Path("path1"), Path("path2")], [Path("path3"), Path("path4")]],
            [["path1", "path2"], ["path3", "path4"]],
        ),
        # Test empty
        ([], []),
    ],
)
def test_ensure_str(
    input_list: list[str] | list[list[str]] | list[Path] | list[list[Path]],
    expected_output: list[str] | list[list[str]],
) -> None:
    """Test the _ensure_str function."""
    output_list = _ensure_str(input_list)
    assert output_list == expected_output


@pytest.mark.parametrize(
    "invalid_input",
    [
        # Test with an invalid input
        {1, 2},
        10,
        10.0,
    ],
)
def test_ensure_str_invalid_input(invalid_input: list[str]) -> None:
    """Test the _ensure_str function with an invalid input."""
    with pytest.raises(ValueError, match="Invalid input:"):
        _ensure_str(invalid_input)  # type: ignore[arg-type]


@pytest.mark.asyncio
async def test_dependencies(
    db_manager: DatabaseManager,
    fnames: list[str] | list[Path],
    socket: zmq.asyncio.Socket,
) -> None:
    """Test the dependencies between learners."""
    db_manager.dependencies = {1: [0]}
    db_manager.create_empty_db()
    db_manager.start()
    await asyncio.sleep(0.1)  # Give it some time to start
    assert db_manager.task is not None
    assert db_manager._db is not None

    # Start a job for learner1
    job_id1, log_fname1, job_name1 = "1000", "log1.log", "job_name1"
    start_message1 = ("start", job_id1, log_fname1, job_name1)
    _index1, _fname1 = db_manager._choose_fname()
    db_manager._confirm_submitted(_index1, job_name1)
    fname1 = await send_message(socket, start_message1)
    assert fname1 == _fname1
    assert fname1 == _ensure_str(fnames[0]), fname1
    e = db_manager._db.get(lambda entry: entry.fname == fname1)
    assert isinstance(e, _DBEntry)
    assert e.job_id == job_id1

    # Try getting a new job
    for _ in range(2):  # try twice
        index, _ = db_manager._choose_fname()
        assert index == -1  # means we are waiting for the dependency to finish

    # Mark the job for learner1 as done
    db_manager._stop_request(fname1)
    entry = db_manager._db.get(lambda entry: entry.fname == fname1)
    assert entry is not None
    assert entry.job_id is None
    assert entry.is_done is True
    assert entry.job_name is None

    # Try getting a new job
    _index2, _fname2 = db_manager._choose_fname()

    # Start a job for learner2
    job_id2, log_fname2, job_name2 = "1001", "log2.log", "job_name2"
    start_message2 = ("start", job_id2, log_fname2, job_name2)
    db_manager._confirm_submitted(_index2, job_name2)
    fname2 = await send_message(socket, start_message2)
    assert fname2 == _fname2
    assert fname2 == _ensure_str(fnames[1])
    e = db_manager._db.get(lambda entry: entry.fname == fname2)
    assert isinstance(e, _DBEntry)
    assert e.job_id == job_id2

    # Stop the job for learner2 using _stop_requests
    db_manager._stop_requests([fname2])

    entry = db_manager._db.get(lambda entry: entry.fname == fname2)
    assert entry is not None
    assert entry.job_id is None, (fname1, fname2)
    assert entry.is_done is True
    assert entry.job_name is None
    with pytest.raises(
        RuntimeError,
        match="Requested a new job but no more learners to run in the database.",
    ):
        db_manager._choose_fname()


@pytest.mark.asyncio
async def test_replace_learner(db_manager: DatabaseManager) -> None:
    """Test replacing a learner in the DatabaseManager."""
    db_manager.create_empty_db()

    # Create a new learner to replace the existing one
    new_learner = adaptive.Learner1D(lambda x: x**2, bounds=(-1, 1))

    # Get the original learner and its fname
    original_learner = db_manager.learners[0]
    original_fname = db_manager.fnames[0]

    # Replace the learner
    db_manager.replace_learner(0, new_learner)

    # Check if the learner was replaced in the list
    assert db_manager.learners[0] is new_learner
    assert db_manager.learners[0] is not original_learner

    # Check if the database entry was updated
    assert db_manager._db is not None
    entry = db_manager._db.get(lambda e: e.fname == _ensure_str(original_fname))
    assert entry is not None
    assert not entry.is_done
    assert not entry.is_pending
    assert entry.job_id is None

    # Check if the cloudpickled file was updated
    cloudpickle_fname = fname_to_learner_fname(original_fname)
    assert cloudpickle_fname.exists()

    # Try to replace a non-existent learner (should raise an IndexError)
    with pytest.raises(IndexError, match="Index out of range"):
        db_manager.replace_learner(len(db_manager.learners), new_learner)

    # Mark the first learner as done
    db_manager._db.update({"is_done": True}, indices=[0])

    # Try to replace a completed learner (should raise a ValueError)
    with pytest.raises(
        ValueError,
        match="Learner at index 0 is already done and cannot be replaced.",
    ):
        db_manager.replace_learner(0, new_learner)



"""Client support for Adaptive Scheduler."""

from __future__ import annotations

import datetime
import json
import logging
import os
import socket
from contextlib import suppress
from typing import TYPE_CHECKING, Any

import psutil
import structlog
import zmq

from adaptive_scheduler.utils import (
    _deserialize,
    _get_npoints,
    _serialize,
    fname_to_learner,
    log_exception,
    sleep_unless_task_is_done,
)

if TYPE_CHECKING:
    import argparse
    import asyncio
    from collections.abc import Callable
    from pathlib import Path

    from adaptive import AsyncRunner, BaseLearner


def _dumps(event_dict: dict[str, Any], **kwargs: Any) -> str:
    """Custom json.dumps to ensure 'event' key is always first in the JSON output."""
    event = event_dict.pop("event", None)
    return json.dumps({"event": event, **event_dict}, **kwargs)


ctx = zmq.Context()
logger = logging.getLogger("adaptive_scheduler.client")
logger.setLevel(logging.INFO)
log = structlog.wrap_logger(
    logger,
    processors=[
        structlog.processors.StackInfoRenderer(),
        structlog.processors.format_exc_info,
        structlog.processors.TimeStamper(fmt="%Y-%m-%d %H:%M.%S", utc=False),
        structlog.processors.JSONRenderer(serializer=_dumps),
    ],
)


def add_log_file_handler(log_fname: str | Path) -> None:  # pragma: no cover
    """Add a file handler to the logger."""
    fh = logging.FileHandler(log_fname)
    logger.addHandler(fh)


def get_learner(
    url: str,
    log_fname: str,
    job_id: str,
    job_name: str,
) -> tuple[BaseLearner, str | list[str], Callable[[], None] | None]:
    """Get a learner from the database (running at `url`).

    This learner's process will be logged in `log_fname`
    and running under `job_id`.

    Parameters
    ----------
    url
        The url of the database manager running via
        (`adaptive_scheduler.server_support.manage_database`).
    log_fname
        The filename of the log-file. Should be passed in the job-script.
    job_id
        The job_id of the process the job. Should be passed in the job-script.
    job_name
        The name of the job. Should be passed in the job-script.

    Returns
    -------
    learner
        Learner that is chosen.
    fname
        The filename of the learner that was chosen.
    initializer
        A function that runs before the process is forked.

    """
    log.info(
        "trying to get learner",
        job_id=job_id,
        log_fname=log_fname,
        job_name=job_name,
    )
    with ctx.socket(zmq.REQ) as socket:
        socket.setsockopt(zmq.LINGER, 0)
        socket.setsockopt(zmq.SNDTIMEO, 300_000)  # timeout after 300s
        socket.connect(url)
        socket.send_serialized(("start", job_id, log_fname, job_name), _serialize)
        log.info("sent start signal, going to wait 300s for a reply.")
        socket.setsockopt(zmq.RCVTIMEO, 300_000)  # timeout after 300s
        reply = socket.recv_serialized(_deserialize)
        log.info("got reply", reply=str(reply))
        if reply is None:
            msg = "No learners to be run."
            exception = RuntimeError(msg)
            log_exception(log, msg, exception)
            raise exception
        if isinstance(reply, Exception):
            log_exception(log, "got an exception", exception=reply)
            raise reply
        fname = reply
        learner, initializer = fname_to_learner(fname, return_initializer=True)
        log.info("got fname and loaded learner")

    log.info("picked a learner")
    return learner, fname, initializer


def tell_done(url: str, fname: str | list[str]) -> None:
    """Tell the database that the learner has reached it's goal.

    Parameters
    ----------
    url
        The url of the database manager running via
        (`adaptive_scheduler.server_support.manage_database`).
    fname
        The filename of the learner that is done.

    """
    log.info("goal reached! 🎉🎊🥳")
    with ctx.socket(zmq.REQ) as socket:
        socket.setsockopt(zmq.LINGER, 0)
        socket.connect(url)
        socket.setsockopt(zmq.SNDTIMEO, 300_000)  # timeout after 300s
        socket.send_serialized(("stop", fname), _serialize)
        socket.setsockopt(zmq.RCVTIMEO, 300_000)  # timeout after 300s
        log.info("sent stop signal, going to wait 300s for a reply", fname=fname)
        socket.recv_serialized(_deserialize)  # Needed because of socket type


def _get_log_entry(runner: AsyncRunner, npoints_start: int) -> dict[str, Any]:
    learner = runner.learner
    info: dict[str, float | str] = {}
    Δt = datetime.timedelta(seconds=runner.elapsed_time())  # noqa: N806, PLC2401
    info["elapsed_time"] = str(Δt)
    info["overhead"] = runner.overhead()
    npoints = _get_npoints(learner)
    if npoints is not None:
        info["npoints"] = npoints
        Δnpoints = npoints - npoints_start  # noqa: N806, PLC2401
        with suppress(ZeroDivisionError):
            # Δt.seconds could be zero if the job is done when starting
            info["npoints/s"] = Δnpoints / Δt.seconds
    with suppress(Exception):
        info["latest_loss"] = learner._cache["loss"]
    with suppress(AttributeError):
        info["nlearners"] = len(learner.learners)
        if "npoints" in info:
            info["npoints/learner"] = info["npoints"] / info["nlearners"]  # type: ignore[operator]
    info["cpu_usage"] = psutil.cpu_percent()
    info["mem_usage"] = psutil.virtual_memory().percent
    for k, v in psutil.cpu_times()._asdict().items():
        info[f"cputimes.{k}"] = v
    return info


def log_now(runner: AsyncRunner, npoints_start: int) -> None:
    """Create a log message now."""
    info = _get_log_entry(runner, npoints_start)
    log.info("current status", **info)


def log_info(runner: AsyncRunner, interval: float = 300) -> asyncio.Task:
    """Log info in the job's logfile, similar to `runner.live_info`.

    Parameters
    ----------
    runner
        Adaptive Runner instance.
    interval
        Time in seconds between log entries.

    """

    async def coro(runner: AsyncRunner, interval: float) -> None:
        log.info(f"started logger on hostname {socket.gethostname()}")  # noqa: G004
        learner = runner.learner
        npoints_start = _get_npoints(learner)
        assert npoints_start is not None
        log.info("npoints at start", npoints=npoints_start)
        while runner.status() == "running":
            if await sleep_unless_task_is_done(runner.task, interval):
                break
            log_now(runner, npoints_start)
        log.info("runner status changed", status=runner.status())
        log.info("current status", **_get_log_entry(runner, npoints_start))

    return runner.ioloop.create_task(coro(runner, interval))


def args_to_env(args: argparse.Namespace, prefix: str = "ADAPTIVE_SCHEDULER_") -> None:
    """Convert parsed arguments to environment variables."""
    env_vars = {}
    for arg, value in vars(args).items():
        if value is not None:
            env_vars[f"{prefix}{arg.upper()}"] = str(value)
    os.environ.update(env_vars)
    log.info("set environment variables", **env_vars)



from __future__ import annotations

import abc
import asyncio
import datetime
import functools
import os
import time
import uuid
from concurrent.futures import Executor, Future
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, NamedTuple

import cloudpickle
from adaptive import SequenceLearner

import adaptive_scheduler

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    from adaptive_scheduler.utils import (
        _DATAFRAME_FORMATS,
        EXECUTOR_TYPES,
        LOKY_START_METHODS,
        GoalTypes,
    )


class AdaptiveSchedulerExecutorBase(Executor):
    _run_manager: adaptive_scheduler.RunManager | None

    @abc.abstractmethod
    def submit(self, fn: Callable[..., Any], /, *args: Any, **kwargs: Any) -> Future:
        """Submit a task to the executor."""

    @abc.abstractmethod
    def finalize(self, *, start: bool = True) -> adaptive_scheduler.RunManager | None:
        """Finalize the executor and return the RunManager.

        Returns None if no learners were submitted.
        """

    def map(  # type: ignore[override]
        self,
        fn: Callable[..., Any],
        /,
        *iterables: Iterable[Any],
        timeout: float | None = None,
        chunksize: int = 1,
    ) -> list[Future]:
        tasks = []
        if timeout is not None:
            msg = "Timeout not implemented"
            raise NotImplementedError(msg)
        if chunksize != 1:
            msg = "Chunksize not implemented"
            raise NotImplementedError(msg)
        for args in zip(*iterables, strict=True):
            task = self.submit(fn, *args)
            tasks.append(task)
        return tasks

    def shutdown(
        self,
        wait: bool = True,  # noqa: FBT001, FBT002
        *,
        cancel_futures: bool = False,
    ) -> None:
        if not wait:
            msg = "Non-waiting shutdown not implemented"
            raise NotImplementedError(msg)
        if cancel_futures:
            msg = "Cancelling futures not implemented"
            raise NotImplementedError(msg)
        if self._run_manager is not None:
            self._run_manager.cancel()


class TaskID(NamedTuple):
    learner_index: int
    sequence_index: int


class SlurmTask(Future):
    """A `Future` that loads the result from a `SequenceLearner`."""

    def __init__(
        self,
        executor: SlurmExecutor,
        task_id: TaskID,
        min_load_interval: float = 1.0,
    ) -> None:
        super().__init__()
        self.executor = executor
        self.task_id = task_id
        self.min_load_interval: float = min_load_interval
        self._last_size: float = 0
        self._load_time: float = 0

        loop = asyncio.get_event_loop()
        self._background_task = loop.create_task(self._background_check())

    async def _background_check(self) -> None:
        """Periodically check if the task is done."""
        while not self.done():
            if self.executor._run_manager is not None:
                self._get()
            await asyncio.sleep(1)

    def _get(self) -> Any | None:  # noqa: PLR0911
        """Updates the state of the task and returns the result if the task is finished."""
        if self.done():
            return super().result(timeout=0)

        func_id, global_index = self.task_id
        try:
            learner_idx, local_index = self.executor._task_mapping[(func_id, global_index)]
        except KeyError as e:
            msg = "Task mapping not found; finalize() must be called first."
            raise RuntimeError(msg) from e
        # Now retrieve the correct learner and filename:
        run_manager = self.executor._run_manager
        assert run_manager is not None, "RunManager not initialized"
        learner = run_manager.learners[learner_idx]
        fname = run_manager.fnames[learner_idx]

        if learner.done():
            result = learner.data[local_index]
            self.set_result(result)
            return result

        assert self.executor._run_manager is not None
        last_load_time = self.executor._run_manager._last_load_time.get(learner_idx, 0)
        now = time.monotonic()
        time_since_last_load = now - last_load_time
        if time_since_last_load < self.min_load_interval:
            return None

        try:
            size = os.path.getsize(fname)  # noqa: PTH202
        except FileNotFoundError:
            return None

        if self._last_size == size:
            return None
        self._last_size = size

        learner.load(fname)
        self._load_time = time.monotonic() - now
        self.min_load_interval = max(1.0, 20.0 * self._load_time)
        self.executor._run_manager._last_load_time[learner_idx] = now

        if local_index in learner.data:
            result = learner.data[local_index]
            self.set_result(result)
            return result
        return None

    @functools.cached_property
    def _learner_and_fname(self) -> tuple[SequenceLearner, str | Path]:
        idx_learner, _ = self.task_id
        run_manager = self.executor._run_manager
        assert run_manager is not None, "RunManager not initialized"
        learner: SequenceLearner = run_manager.learners[idx_learner]  # type: ignore[index]
        fname = run_manager.fnames[idx_learner]
        return learner, fname

    def result(self, timeout: float | None = None) -> Any:
        """Return the result of the future if available.

        Since this is an async task, this method will only return if the result
        is immediately available. Use `await task` to wait for the result.
        """
        if timeout is not None:
            msg = "Timeout not implemented for SLURMTask"
            raise NotImplementedError(msg)

        if self.executor._run_manager is None:
            msg = "RunManager not initialized. Call finalize() first."
            raise RuntimeError(msg)

        # Do one check
        self._get()

        if not self.done():
            msg = (
                "Result not immediately available. "
                "Use 'await task' to wait for the result asynchronously."
            )
            raise RuntimeError(msg)

        return super().result(timeout=0)  # timeout=0 makes it non-blocking

    def cancel(self) -> bool:
        """Cancel the future and its background task."""
        self._background_task.cancel()
        return super().cancel()

    def __repr__(self) -> str:
        if not self.done():
            self._get()
        return f"SLURMTask(task_id={self.task_id}, state={self._state})"

    def __str__(self) -> str:
        return self.__repr__()

    def __await__(self) -> Any:
        """Allow using 'await task' to wait for the result."""
        return asyncio.wrap_future(self).__await__()


def _uuid_with_datetime() -> str:
    """Return a UUID with the current datetime."""
    # YYYYMMDD-HHMMSS-UUID
    return f"{datetime.datetime.now().strftime('%Y%m%d-%H%M%S')}-{uuid.uuid4().hex}"  # noqa: DTZ005


class _SerializableFunctionSplatter:
    def __init__(self, func: Callable[..., Any]) -> None:
        self.func = func

    def __call__(self, args: Any) -> Any:
        return self.func(*args)

    def __getstate__(self) -> bytes:
        return cloudpickle.dumps(self.func)

    def __setstate__(self, state: bytes) -> None:
        self.func = cloudpickle.loads(state)


@dataclass
class SlurmExecutor(AdaptiveSchedulerExecutorBase):
    """An executor that runs jobs on SLURM.

    Similar to `concurrent.futures.Executor`, but for SLURM.
    A key difference is that ``submit()`` returns a `SLURMTask` instead of a `Future`
    and that ``finalize()`` must be called in order to start the jobs.

    Parameters
    ----------
    name
        The name of the job.
    folder
        The folder to save the adaptive_scheduler files such as logs, database,
        ``.sbatch``, pickled tasks, and results files in. If the folder exists and has
        results, the results will be loaded!
    partition
        The partition to use. If None, then the default partition will be used.
        (The one marked with a * in `sinfo`). Use
        `adaptive_scheduler.scheduler.slurm_partitions` to see the
        available partitions.
    nodes
        The number of nodes to use.
    cores_per_node
        The number of cores per node to use. If None, then all cores on the partition
        will be used.
    num_threads
        The number of threads to use.
    exclusive
        Whether to use exclusive nodes, adds ``"--exclusive"`` if True.
    executor_type
        The executor that is used, by default `concurrent.futures.ProcessPoolExecutor` is used.
        One can use ``"ipyparallel"``, ``"dask-mpi"``, ``"mpi4py"``,
        ``"loky"``, ``"sequential"``, or ``"process-pool"``.
    extra_scheduler
        Extra ``#SLURM`` (depending on scheduler type)
        arguments, e.g. ``["--exclusive=user", "--time=1"]`` or a tuple of lists,
        e.g. ``(["--time=10"], ["--time=20"]])`` for two jobs.
    goal
        The goal passed to the `adaptive.Runner`. Note that this function will
        be serialized and pasted in the ``job_script``. Can be a smart-goal
        that accepts
        ``Callable[[adaptive.BaseLearner], bool] | float | datetime | timedelta | None``.
        See `adaptive_scheduler.utils.smart_goal` for more information.
    check_goal_on_start
        Checks whether a learner is already done. Only works if the learner is loaded.
    runner_kwargs
        Extra keyword argument to pass to the `adaptive.Runner`. Note that this dict
        will be serialized and pasted in the ``job_script``.
    url
        The url of the database manager, with the format
        ``tcp://ip_of_this_machine:allowed_port.``. If None, a correct url will be chosen.
    save_interval
        Time in seconds between saving of the learners.
    log_interval
        Time in seconds between log entries.
    job_manager_interval
        Time in seconds between checking and starting jobs.
    kill_interval
        Check for `kill_on_error` string inside the log-files every `kill_interval` seconds.
    kill_on_error
        If ``error`` is a string and is found in the log files, the job will
        be cancelled and restarted. If it is a callable, it is applied
        to the log text. Must take a single argument, a list of
        strings, and return True if the job has to be killed, or
        False if not. Set to None if no `KillManager` is needed.
    overwrite_db
        Overwrite the existing database.
    job_manager_kwargs
        Keyword arguments for the `JobManager` function that aren't set in ``__init__`` here.
    kill_manager_kwargs
        Keyword arguments for the `KillManager` function that aren't set in ``__init__`` here.
    loky_start_method
        Loky start method, by default "loky".
    cleanup_first
        Cancel all previous jobs generated by the same RunManager and clean logfiles.
    save_dataframe
        Whether to periodically save the learner's data as a `pandas.DataFame`.
    dataframe_format
        The format in which to save the `pandas.DataFame`. See the type hint for the options.
    max_log_lines
        The maximum number of lines to display in the log viewer widget.
    max_fails_per_job
        Maximum number of times that a job can fail. This is here as a fail switch
        because a job might fail instantly because of a bug inside your code.
        The job manager will stop when
        ``n_jobs * total_number_of_jobs_failed > max_fails_per_job`` is true.
    max_simultaneous_jobs
        Maximum number of simultaneously running jobs. By default no more than 500
        jobs will be running. Keep in mind that if you do not specify a ``runner.goal``,
        jobs will run forever, resulting in the jobs that were not initially started
        (because of this `max_simultaneous_jobs` condition) to not ever start.
    quiet
        Whether to show a progress bar when creating learner files.
    extra_run_manager_kwargs
        Extra keyword arguments to pass to the `RunManager`.
    extra_scheduler_kwargs
        Extra keyword arguments to pass to the `adaptive_scheduler.scheduler.SLURM`.
    size_per_learner
        The size of each learner. If None, the whole sequence is passed to the learner.

    """

    # Same as slurm_run, except it has no learners, fnames, dependencies and initializers.

    # slurm_run: Specific to slurm_run
    name: str = "adaptive-scheduler"
    folder: str | Path | None = None  # `slurm_run` defaults to None
    # slurm_run: SLURM scheduler arguments
    partition: str | tuple[str | Callable[[], str], ...] | None = None
    nodes: int | tuple[int | None | Callable[[], int | None], ...] | None = 1
    cores_per_node: int | tuple[int | None | Callable[[], int | None], ...] | None = (
        1  # `slurm_run` defaults to `None`
    )
    num_threads: int | tuple[int | Callable[[], int], ...] = 1
    exclusive: bool | tuple[bool | Callable[[], bool], ...] = False
    executor_type: EXECUTOR_TYPES | tuple[EXECUTOR_TYPES | Callable[[], EXECUTOR_TYPES], ...] = (
        "process-pool"
    )
    extra_scheduler: list[str] | tuple[list[str] | Callable[[], list[str]], ...] | None = None
    # slurm_run: Same as RunManager below (except dependencies and initializers)
    goal: GoalTypes | None = None
    check_goal_on_start: bool = True
    runner_kwargs: dict | None = None
    url: str | None = None
    save_interval: float = 300
    log_interval: float = 300
    job_manager_interval: float = 60
    kill_interval: float = 60
    kill_on_error: str | Callable[[list[str]], bool] | None = "srun: error:"
    overwrite_db: bool = True
    job_manager_kwargs: dict[str, Any] | None = None
    kill_manager_kwargs: dict[str, Any] | None = None
    loky_start_method: LOKY_START_METHODS = "loky"
    cleanup_first: bool = True
    save_dataframe: bool = False  # `slurm_run` defaults to `True`
    dataframe_format: _DATAFRAME_FORMATS = "pickle"
    max_log_lines: int = 500
    max_fails_per_job: int = 50
    max_simultaneous_jobs: int = 100
    quiet: bool = True  # `slurm_run` defaults to `False`
    # slurm_run: RunManager arguments
    extra_run_manager_kwargs: dict[str, Any] | None = None
    extra_scheduler_kwargs: dict[str, Any] | None = None
    # Internal
    size_per_learner: int | None = None
    _sequences: dict[Callable[..., Any], list[Any]] = field(default_factory=dict)
    _sequence_mapping: dict[Callable[..., Any], int] = field(default_factory=dict)
    _disk_func_mapping: dict[Callable[..., Any], _DiskFunction] = field(default_factory=dict)
    _run_manager: adaptive_scheduler.RunManager | None = None
    _task_mapping: dict[tuple[int, int], tuple[int, int]] = field(default_factory=dict)

    def __post_init__(self) -> None:
        if self.folder is None:
            self.folder = Path.cwd() / ".adaptive_scheduler" / _uuid_with_datetime()  # type: ignore[operator]
        else:
            self.folder = Path(self.folder)

    def submit(self, fn: Callable[..., Any], /, *args: Any, **kwargs: Any) -> SlurmTask:
        if kwargs:
            msg = "Keyword arguments are not supported"
            raise ValueError(msg)
        if fn not in self._sequence_mapping:
            self._sequence_mapping[fn] = len(self._sequence_mapping)
            assert fn not in self._disk_func_mapping
            assert isinstance(self.folder, Path)
            self._disk_func_mapping[fn] = _DiskFunction(
                fn,
                self.folder / f"{_name(fn)}-{uuid.uuid4().hex}.pickle",
            )
        sequence = self._sequences.setdefault(fn, [])
        i = len(sequence)
        sequence.append(args)
        task_id = TaskID(self._sequence_mapping[fn], i)
        return SlurmTask(self, task_id)

    def _to_learners(
        self,
    ) -> tuple[
        list[SequenceLearner],
        list[Path],
        dict[tuple[int, int], tuple[int, int]],
    ]:
        learners = []
        fnames = []
        task_mapping = {}
        learner_idx = 0
        for func, args_list in self._sequences.items():
            func_id = self._sequence_mapping[func]
            # Chunk the sequence if size_per_learner is set; otherwise one chunk.
            if self.size_per_learner is not None:
                chunked_args = [
                    args_list[i : i + self.size_per_learner]
                    for i in range(0, len(args_list), self.size_per_learner)
                ]
            else:
                chunked_args = [args_list]

            global_index = 0  # global index for tasks of this function
            for chunk in chunked_args:
                # Map each task in the chunk: global index -> (current learner, local index)
                for local_index in range(len(chunk)):
                    task_mapping[(func_id, global_index)] = (learner_idx, local_index)
                    global_index += 1

                disk_func = self._disk_func_mapping[func]
                ser_func = _SerializableFunctionSplatter(disk_func)
                learner = SequenceLearner(ser_func, chunk)
                learners.append(learner)
                name = _name(func)
                assert isinstance(self.folder, Path)
                fnames.append(self.folder / f"{name}-{learner_idx}-{uuid.uuid4().hex}.pickle")
                learner_idx += 1
        return learners, fnames, task_mapping

    def finalize(self, *, start: bool = True) -> adaptive_scheduler.RunManager | None:
        if self._run_manager is not None:
            msg = "RunManager already initialized. Create a new SlurmExecutor instance."
            raise RuntimeError(msg)
        learners, fnames, self._task_mapping = self._to_learners()
        if not learners:
            return None
        assert self.folder is not None
        self._run_manager = adaptive_scheduler.slurm_run(
            learners=learners,
            fnames=fnames,
            # Specific to slurm_run
            name=self.name,
            folder=self.folder,
            # SLURM scheduler arguments
            partition=self.partition,
            nodes=self.nodes,
            cores_per_node=self.cores_per_node,
            num_threads=self.num_threads,
            exclusive=self.exclusive,
            executor_type=self.executor_type,
            extra_scheduler=self.extra_scheduler,
            # Same as RunManager below (except job_name, move_old_logs_to, and db_fname)
            goal=self.goal,
            check_goal_on_start=self.check_goal_on_start,
            runner_kwargs=self.runner_kwargs,
            url=self.url,
            save_interval=self.save_interval,
            log_interval=self.log_interval,
            job_manager_interval=self.job_manager_interval,
            kill_interval=self.kill_interval,
            kill_on_error=self.kill_on_error,
            overwrite_db=self.overwrite_db,
            job_manager_kwargs=self.job_manager_kwargs,
            kill_manager_kwargs=self.kill_manager_kwargs,
            loky_start_method=self.loky_start_method,
            cleanup_first=self.cleanup_first,
            save_dataframe=self.save_dataframe,
            dataframe_format=self.dataframe_format,
            max_log_lines=self.max_log_lines,
            max_fails_per_job=self.max_fails_per_job,
            max_simultaneous_jobs=self.max_simultaneous_jobs,
            quiet=self.quiet,
            # RunManager arguments
            extra_run_manager_kwargs=self.extra_run_manager_kwargs,
            extra_scheduler_kwargs=self.extra_scheduler_kwargs,
        )
        if start:
            self._run_manager.start()
        return self._run_manager

    def cleanup(self) -> None:
        assert self._run_manager is not None
        self._run_manager.cleanup(remove_old_logs_folder=True)

    def new(self, update: dict[str, Any] | None = None) -> SlurmExecutor:
        """Create a new SlurmExecutor with the same parameters."""
        data = asdict(self)
        data["_run_manager"] = None
        data["_sequences"] = {}
        data["_sequence_mapping"] = {}
        if update is not None:
            data.update(update)
        return SlurmExecutor(**data)


def _name(func: Callable[..., Any]) -> str:
    return func.__name__ if hasattr(func, "__name__") else "func"


class _DiskFunction:
    def __init__(self, func: Callable[..., Any], fname: str | Path) -> None:
        self.fname = Path(fname)
        self.fname.parent.mkdir(parents=True, exist_ok=True)
        with self.fname.open("wb") as f:
            cloudpickle.dump(func, f)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.func(*args, **kwargs)

    @functools.cached_property
    def func(self) -> Callable[..., Any]:
        return cloudpickle.loads(self.fname.read_bytes())
